from dataclasses import dataclass
from .tipos_comuns import Cabecalho
from lxml.etree import Element, SubElement, tostring, indent
from abstra_notas.validacoes.cnpj import normalizar_cnpj


@dataclass
class PedidoConsultaCNPJ:
    cabecalho: Cabecalho
    cnpj_contribuinte: str

    def __post_init__(self):
        self.cnpj_contribuinte = normalizar_cnpj(self.cnpj_contribuinte)

    def gerar_xml(self):
        pedido_consulta_cnpj = Element(
            "PedidoConsultaCNPJ", xmlns="http://www.prefeitura.sp.gov.br/nfe"
        )
        cabecalho = self.cabecalho.gerar_xml()
        pedido_consulta_cnpj.append(cabecalho)
        cnpj_contribuinte = Element("CNPJContribuinte", xmlns="")
        SubElement(cnpj_contribuinte, "CNPJ").text = self.cnpj_contribuinte
        pedido_consulta_cnpj.append(cnpj_contribuinte)
        return pedido_consulta_cnpj

    @staticmethod
    def parse_xml(element: Element) -> "PedidoConsultaCNPJ":
        cabecalho = Cabecalho.parse_xml(element.find("Cabecalho"))
        cnpj_contribuinte = element.find("CNPJContribuinte").find("CNPJ").text
        return PedidoConsultaCNPJ(
            cabecalho=cabecalho, cnpj_contribuinte=cnpj_contribuinte
        )

    def gerar_string(self) -> str:
        xml = self.gerar_xml()
        indent(xml, space="  ")
        return tostring(xml, method="xml", encoding="unicode")



from unittest import TestCase
from .consulta_cnpj_pedido import PedidoConsultaCNPJ
from .tipos_comuns import Cabecalho, CPFCNPJRemetente
from abstra_notas.assinatura import Assinador, gerar_pfx_teste
from lxml.etree import XMLSchema
from pathlib import Path
from tempfile import NamedTemporaryFile


class TestConsultaCnpjPedido(TestCase):
    def test_xml(self):
        pedido = PedidoConsultaCNPJ(
            cabecalho=Cabecalho(
                cpf_cnpj_remetente=CPFCNPJRemetente(
                    tipo="CNPJ", numero="99999997000100"
                )
            ),
            cnpj_contribuinte="60.701.190/0001-04",
        )

        schema = XMLSchema(
            file=Path(__file__).parent / "xsds" / "PedidoConsultaCNPJ_v01.xsd"
        )
        xml = pedido.gerar_xml()
        with NamedTemporaryFile() as f:
            pfx_path = Path(f.name)
            password = "123456"
            gerar_pfx_teste(path=pfx_path, password=password)
            xml_assinado = Assinador(pfx_path=pfx_path, pfx_password=password).assinar(
                xml
            )
            schema.assertValid(xml_assinado)



from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.serialization.pkcs12 import (
    load_key_and_certificates,
)
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID
from cryptography import x509
from os import getenv
from pathlib import Path
from tempfile import NamedTemporaryFile
from lxml.etree import Element, tostring, fromstring
import xmlsec
import datetime


class Assinador:
    pfx_path: Path
    pfx_password: str
    cert_pem_bytes: bytes
    private_key_pem_bytes: bytes

    def __init__(
        self,
        pfx_path: Path = Path(getenv("NFSE_PFX_PATH")),
        pfx_password: str = getenv("NFSE_PFX_PASSWORD"),
    ):
        self.pfx_path = pfx_path
        self.pfx_password = pfx_password

        with open(self.pfx_path, "rb") as f:
            pfx_data = f.read()

        private_key, certificate, _ = load_key_and_certificates(
            pfx_data, self.pfx_password.encode(), backend=default_backend()
        )

        private_key_pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )

        cert_pem = certificate.public_bytes(serialization.Encoding.PEM)

        self.cert_pem_bytes = cert_pem
        self.private_key_pem_bytes = private_key_pem

    @property
    def cert_pem_file(self):
        file = NamedTemporaryFile()
        file.write(self.cert_pem_bytes)
        file.seek(0)
        return file

    @property
    def private_key_pem_file(self):
        file = NamedTemporaryFile()
        file.write(self.private_key_pem_bytes)
        file.seek(0)
        return file

    def assinar_xml(self, element: Element) -> Element:
        with self.private_key_pem_file as private_key_pem_file, self.cert_pem_file as cert_pem_file:
            element = fromstring(tostring(element, encoding=str))
            key = xmlsec.Key.from_file(
                private_key_pem_file.name,
                format=xmlsec.constants.KeyDataFormatPem,
                password=self.pfx_password,
            )
            signature_node: Element = xmlsec.template.create(
                element,
                c14n_method=xmlsec.constants.TransformInclC14N,
                sign_method=xmlsec.constants.TransformRsaSha1,
            )
            element.append(signature_node)
            ref = xmlsec.template.add_reference(
                signature_node, xmlsec.constants.TransformSha1, uri=""
            )
            xmlsec.template.add_transform(ref, xmlsec.constants.TransformEnveloped)
            xmlsec.template.add_transform(ref, xmlsec.constants.TransformInclC14N)
            key_info = xmlsec.template.ensure_key_info(signature_node)
            xmlsec.template.add_x509_data(key_info)
            ctx = xmlsec.SignatureContext()
            ctx.key = key
            ctx.key.load_cert_from_file(
                cert_pem_file.name, xmlsec.constants.KeyDataFormatPem
            )
            ctx.sign(signature_node)
            print(tostring(element, encoding=str))
            return element

    def assinar_bytes_rsa_sh1(self, data: bytes) -> bytes:
        private_key = serialization.load_pem_private_key(
            self.private_key_pem_bytes,
            password=None,
            backend=default_backend(),
        )

        signature = private_key.sign(
            data,
            padding.PKCS1v15(),
            hashes.SHA1(),
        )
        return signature


class AssinadorMock:
    def assinar_xml(self, element: Element) -> Element:
        return element

    def assinar_bytes_rsa_sh1(self, data: bytes) -> bytes:
        return data



from .envio_rps import PedidoEnvioRPS, RetornoEnvioRPS
from .consulta_cnpj import PedidoConsultaCNPJ, RetornoConsultaCNPJ
from .cliente import Cliente

__all__ = [
    "PedidoEnvioRPS",
    "RetornoEnvioRPS",
    "PedidoConsultaCNPJ",
    "RetornoConsultaCNPJ",
    "Cliente",
]



def normalizar_cep(cep: str) -> str:
    cep_normalizado = cep.replace("-", "").replace(".", "")
    if len(cep_normalizado) < 8:
        cep_normalizado = "0" * (8 - len(cep_normalizado)) + cep_normalizado
    assert len(cep_normalizado) == 8, "CEP deve ter 8 dígitos"
    return cep_normalizado



from dataclasses import dataclass
from typing import Literal, Tuple
from lxml.etree import Element, fromstring
import base64
from abstra_notas.validacoes.email import validar_email
from abstra_notas.validacoes.cidades import validar_codigo_cidade, normalizar_uf
from abstra_notas.validacoes.cpf import normalizar_cpf
from abstra_notas.validacoes.cnpj import normalizar_cnpj
from .codigos_de_servico import codigos_de_servico_validos
from datetime import date
from jinja2 import Template
from pathlib import Path
from .pedido import Pedido
from .retorno import Retorno
from abstra_notas.assinatura import Assinador


@dataclass
class RetornoEnvioRPS(Retorno):
    sucesso: bool
    chave_nfe_inscricao_prestador: str
    chave_nfe_numero_nfe: str
    chave_nfe_codigo_verificacao: str
    chave_rps_inscricao_prestador: str
    chave_rps_serie_rps: str
    chave_rps_numero_rps: str

    @staticmethod
    def ler_xml(xml: Element) -> "RetornoEnvioRPS":
        return RetornoEnvioRPS(
            sucesso=xml.find(".//Sucesso").text == "true",
            chave_nfe_inscricao_prestador=xml.find(".//InscricaoPrestador").text,
            chave_nfe_codigo_verificacao=xml.find(".//CodigoVerificacao").text,
            chave_nfe_numero_nfe=xml.find(".//NumeroNFe").text,
            chave_rps_inscricao_prestador=xml.find(".//InscricaoPrestador").text,
            chave_rps_numero_rps=xml.find(".//NumeroRPS").text,
            chave_rps_serie_rps=xml.find(".//SerieRPS").text,
        )


@dataclass
class PedidoEnvioRPS(Pedido):
    remetente: Tuple[Literal["CPF", "CNPJ"], str]
    inscricao_prestador: str
    serie_rps: str
    numero_rps: int
    tipo_rps: Literal["RPS", "RPS-M", "RPS-C"]
    data_emissao: date
    status_rps: Literal["N", "C"]
    tributacao_rps: Literal["T", "F", "A", "B", "D", "M", "N", "R", "S", "X", "V", "P"]
    valor_servicos_centavos: int
    valor_deducoes_centavos: int
    valor_pis_centavos: int
    valor_cofins_centavos: int
    valor_inss_centavos: int
    valor_ir_centavos: int
    valor_csll_centavos: int
    codigo_servico: int
    aliquota_servicos: float
    iss_retido: Literal["true", "false"]
    tomador: Tuple[Literal["CPF", "CNPJ"], str]
    razao_social_tomador: str
    endereco_tipo_logradouro: str
    endereco_logradouro: str
    endereco_numero: str
    endereco_complemento: str
    endereco_bairro: str
    endereco_cidade: int
    endereco_uf: str
    endereco_cep: str
    email_tomador: str
    discriminacao: str

    def __post_init__(self):
        if self.tomador[0] == "CPF":
            self.tomador = ("CPF", normalizar_cpf(self.tomador[1]))
        if self.tomador[0] == "CNPJ":
            self.tomador = ("CNPJ", normalizar_cnpj(self.tomador[1]))
        self.endereco_uf = normalizar_uf(self.endereco_uf)
        assert validar_codigo_cidade(
            self.endereco_cidade
        ), f"Código de cidade inválido: {self.endereco_cidade}"
        assert (
            self.aliquota_servicos >= 0 and self.aliquota_servicos <= 1
        ), "A alíquota de serviços deve ser um valor entre 0 e 1"
        assert (
            self.codigo_servico in codigos_de_servico_validos
        ), f"Código de serviço inválido, os códigos válidos são: {codigos_de_servico_validos}"
        assert validar_email(
            self.email_tomador
        ), f"Email do tomador com formato inválido: {self.email_tomador}"
        assert isinstance(
            self.valor_servicos_centavos, int
        ), "O valor de serviços deve ser um valor decimal"
        assert isinstance(
            self.valor_deducoes_centavos, int
        ), "O valor de deduções deve ser um valor decimal"
        assert isinstance(
            self.valor_pis_centavos, int
        ), "O valor de PIS deve ser um valor decimal"
        assert isinstance(
            self.valor_cofins_centavos, int
        ), "O valor de COFINS deve ser um valor decimal"
        assert isinstance(
            self.valor_inss_centavos, int
        ), "O valor de INSS deve ser um valor decimal"
        assert isinstance(
            self.valor_ir_centavos, int
        ), "O valor de IR deve ser um valor decimal"
        assert isinstance(
            self.valor_csll_centavos, int
        ), "O valor de CSLL deve ser um valor decimal"
        assert (
            self.valor_servicos_centavos >= 0
        ), "O valor de serviços deve ser maior ou igual a zero"
        assert (
            self.valor_deducoes_centavos >= 0
        ), "O valor de deduções deve ser maior ou igual a zero"
        assert (
            self.valor_pis_centavos >= 0
        ), "O valor de PIS deve ser maior ou igual a zero"
        assert (
            self.valor_cofins_centavos >= 0
        ), "O valor de COFINS deve ser maior ou igual a zero"
        assert (
            self.valor_inss_centavos >= 0
        ), "O valor de INSS deve ser maior ou igual a zero"
        assert (
            self.valor_ir_centavos >= 0
        ), "O valor de IR deve ser maior ou igual a zero"
        assert (
            self.valor_csll_centavos >= 0
        ), "O valor de CSLL deve ser maior ou igual a zero"
        assert (
            self.valor_servicos_centavos
            - self.valor_deducoes_centavos
            - self.valor_pis_centavos
            - self.valor_cofins_centavos
            - self.valor_inss_centavos
            - self.valor_ir_centavos
            - self.valor_csll_centavos
            >= 0
        ), "A soma dos valores não pode ser negativa"

    def gerar_xml(self, assinador: Assinador) -> Element:
        xml = self.template.render(
            **self.__dict__, assinatura=self.assinatura(assinador)
        )

        return fromstring(xml)

    @property
    def nome_metodo(self):
        return "EnvioRPS"

    def assinatura(self, assinador: Assinador) -> str:
        template = ""
        template += str(self.inscricao_prestador)
        template += self.serie_rps.upper()
        template += str(self.numero_rps).zfill(12)
        template += self.data_emissao.strftime("%Y%m%d").upper()
        template += self.tributacao_rps
        template += self.status_rps
        template += self.iss_retido == "true" and "S" or "N"
        template += str(self.valor_servicos_centavos).zfill(15)
        template += str(self.valor_deducoes_centavos).zfill(15)
        template += str(self.codigo_servico).zfill(5)
        if self.tomador[0] == "CPF":
            template += "1"
        elif self.tomador[0] == "CNPJ":
            template += "2"
        else:
            template += "3"
        template += (
            self.tomador[1].replace(".", "").replace("-", "").replace("/", "").zfill(14)
        )

        template_bytes = template.encode("ascii")

        signed_template = assinador.assinar_bytes_rsa_sh1(template_bytes)
        return base64.b64encode(signed_template).decode("ascii")

    @property
    def classe_retorno(self):
        return RetornoEnvioRPS



from ....assinatura import Assinador
from zeep import Client, Transport
from requests import Session
from .pedido import Pedido
from lxml.etree import tostring
from pathlib import Path
from .retorno import Retorno


class Cliente:
    assinador: Assinador

    def __init__(self, caminho_pfx: Path, senha_pfx: str):
        self.assinador = Assinador(caminho_pfx, senha_pfx)

    def executar(self, pedido: Pedido) -> Retorno:
        url = "https://nfe.prefeitura.sp.gov.br/ws/lotenfe.asmx?WSDL"
        xml = pedido.gerar_xml(self.assinador)
        with self.assinador.cert_pem_file as cert_pem_file, self.assinador.private_key_pem_file as private_key_pem_file:
            cert = (cert_pem_file.name, private_key_pem_file.name)
            session = Session()
            session.cert = cert
            transport = Transport(session=session, cache=None)
            client = Client(url, transport=transport)
            signed_xml = self.assinador.assinar_xml(xml)

            retorno = getattr(client.service, pedido.nome_metodo)(
                1, tostring(signed_xml, encoding=str)
            )
            return pedido.classe_retorno.ler_xml(retorno)



from unittest import TestCase
from .envio_rps import PedidoEnvioRPS
from pathlib import Path
from lxml.etree import parse, XMLSchema, tostring
from datetime import date
import re
from abstra_notas.assinatura import AssinadorMock


class EnvioTest(TestCase):
    def test_exemplo(self):
        self.maxDiff = None
        exemplo_path = Path(__file__).parent / "exemplos" / "PedidoEnvioRPS.xml"
        exemplo_xml = parse(exemplo_path)

        pedido = PedidoEnvioRPS(
            aliquota_servicos=0.05,
            codigo_servico=7617,
            data_emissao=date(2015, 1, 20),
            discriminacao="Desenvolvimento de Web Site Pessoal.",
            email_tomador="tomador@teste.com.br",
            endereco_bairro="Bela Vista",
            endereco_cep="1310100",
            endereco_cidade=3550308,
            endereco_complemento="Cj 35",
            endereco_logradouro="Paulista",
            endereco_numero="100",
            endereco_tipo_logradouro="Av",
            endereco_uf="SP",
            inscricao_prestador="39616924",
            iss_retido="false",
            numero_rps=4105,
            razao_social_tomador="TOMADOR PF",
            remetente=("CNPJ", "99999997000100"),
            serie_rps="BB",
            status_rps="N",
            tipo_rps="RPS-M",
            tomador=("CPF", "12345678909"),
            tributacao_rps="T",
            valor_cofins_centavos=10,
            valor_csll_centavos=10,
            valor_deducoes_centavos=5000,
            valor_inss_centavos=10,
            valor_ir_centavos=10,
            valor_pis_centavos=10,
            valor_servicos_centavos=20500,
        )
        pedido_xml = pedido.gerar_xml(assinador=AssinadorMock())
        pedido_str: str = tostring(pedido_xml, encoding=str)
        exemplo_str: str = tostring(exemplo_xml, encoding=str)
        pattern = r"<Assinador>.*</Assinador>"
        pedido_str = re.sub(pattern, "", pedido_str)
        exemplo_str = re.sub(pattern, "", exemplo_str)

        self.assertEqual(pedido_str, exemplo_str)
        schema = XMLSchema(
            file=Path(__file__).parent / "xsds" / "PedidoEnvioRPS_v01.xsd"
        )
        schema.assertValid(pedido_xml)
